////////////////////////////////////////////////////////
//
// GEM - Graphics Environment for Multimedia
//
// zmoelnig@iem.kug.ac.at
//
// Implementation file
//
//    Copyright (c) 1997-1998 Mark Danks.
//    Copyright (c) Günther Geiger.
//    Copyright (c) 2001-2002 IOhannes m zmoelnig. forum::für::umläute. IEM
//    For information on usage and redistribution, and for a DISCLAIMER OF ALL
//    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.
//
/////////////////////////////////////////////////////////

////////////////////////
// videoKINECT Plugin for pix_video by Matthias Kronlachner
///////////////////////
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "videoKINECT.h"
using namespace gem;

#include "Gem/RTE.h"
#include "Base/GemException.h"

#include <inttypes.h>

REGISTER_VIDEOFACTORY("kinect", videoKINECT);

class Mutex {
public:
	Mutex() {
		pthread_mutex_init( &m_mutex, NULL );
	}
	void lock() {
		pthread_mutex_lock( &m_mutex );
	}
	void unlock() {
		pthread_mutex_unlock( &m_mutex );
	}
private:
	pthread_mutex_t m_mutex;
};

/* thanks to Yoda---- from IRC */
class MyFreenectDevice : public Freenect::FreenectDevice {
public:
	MyFreenectDevice(freenect_context *_ctx, int _index)
		: Freenect::FreenectDevice(_ctx, _index), m_buffer_depth(FREENECT_VIDEO_RGB_SIZE),m_buffer_video(FREENECT_VIDEO_RGB_SIZE), m_gamma(2048), m_new_rgb_frame(false), m_new_depth_frame(false)
	{
		for( unsigned int i = 0 ; i < 2048 ; i++) {
			float v = i/2048.0;
			v = std::pow(v, 3)* 6;
			m_gamma[i] = v*6*256;
		}
	}
	//~MyFreenectDevice(){}
	// Do not call directly even in child
	void VideoCallback(void* _rgb, uint32_t timestamp) {
		m_rgb_mutex.lock();
		uint8_t* rgb = static_cast<uint8_t*>(_rgb);
		std::copy(rgb, rgb+getVideoBufferSize(), m_buffer_video.begin());
		m_new_rgb_frame = true;
		m_rgb_mutex.unlock();
	};
	// Do not call directly even in child
	void DepthCallback(void* _depth, uint32_t timestamp) {
		m_depth_mutex.lock();
		uint16_t* depth = static_cast<uint16_t*>(_depth);
		for( unsigned int i = 0 ; i < FREENECT_FRAME_PIX ; i++) {
			int pval = m_gamma[depth[i]];
			int lb = pval & 0xff;
			switch (pval>>8) {
			case 0:																					// Changed for RGBA (4* instead of 3*)
				m_buffer_depth[4*i+0] = 255;
				m_buffer_depth[4*i+1] = 255-lb;
				m_buffer_depth[4*i+2] = 255-lb;
				break;
			case 1:
				m_buffer_depth[4*i+0] = 255;
				m_buffer_depth[4*i+1] = lb;
				m_buffer_depth[4*i+2] = 0;
				break;
			case 2:
				m_buffer_depth[4*i+0] = 255-lb;
				m_buffer_depth[4*i+1] = 255;
				m_buffer_depth[4*i+2] = 0;
				break;
			case 3:
				m_buffer_depth[4*i+0] = 0;
				m_buffer_depth[4*i+1] = 255;
				m_buffer_depth[4*i+2] = lb;
				break;
			case 4:
				m_buffer_depth[4*i+0] = 0;
				m_buffer_depth[4*i+1] = 255-lb;
				m_buffer_depth[4*i+2] = 255;
				break;
			case 5:
				m_buffer_depth[4*i+0] = 0;
				m_buffer_depth[4*i+1] = 0;
				m_buffer_depth[4*i+2] = 255-lb;
				break;
			default:
				m_buffer_depth[4*i+0] = 0;
				m_buffer_depth[4*i+1] = 0;
				m_buffer_depth[4*i+2] = 0;
				break;
			}
		}
		m_new_depth_frame = true;
		m_depth_mutex.unlock();
	}
	bool getRGB(std::vector<uint8_t> &buffer) {
		m_rgb_mutex.lock();
		if(m_new_rgb_frame) {
			buffer.swap(m_buffer_video);
			m_new_rgb_frame = false;
			m_rgb_mutex.unlock();
			return true;
		} else {
			m_rgb_mutex.unlock();
			return false;
		}
	}

	bool getDepth(std::vector<uint8_t> &buffer) {
		m_depth_mutex.lock();
		if(m_new_depth_frame) {
			buffer.swap(m_buffer_depth);
			m_new_depth_frame = false;
			m_depth_mutex.unlock();
			return true;
		} else {
			m_depth_mutex.unlock();
			return false;
		}
	}

private:
	std::vector<uint8_t> m_buffer_depth;
	std::vector<uint8_t> m_buffer_video;
	std::vector<uint16_t> m_gamma;
	Mutex m_rgb_mutex;
	Mutex m_depth_mutex;
	bool m_new_rgb_frame;
	bool m_new_depth_frame;
};

Freenect::Freenect freenect;
MyFreenectDevice* device;
freenect_video_format requested_format(FREENECT_VIDEO_RGB);


/////////////////////////////////////////////////////////
//
// videoKINECT
//
/////////////////////////////////////////////////////////
// Constructor
//
/////////////////////////////////////////////////////////
videoKINECT :: videoKINECT() : video("kinect",0)
{
  
  m_width=640;
  m_height=480;
  x_angle = 0;
  x_led = 1;
  channel=0; // RGB Standard
  video_format=FREENECT_VIDEO_RGB;
	depth_format=FREENECT_DEPTH_11BIT;
	
  provide("freenect");
  
}

////////////////////////////////////////////////////////
// Destructor
//
////////////////////////////////////////////////////////
videoKINECT :: ~videoKINECT(){
  close();
  
}
////////////////////////////////////////////////////////
// render
//
/////////////////////////////////////////////////////////


// this reads the data that was captured and returns it within a pixBlock

	static std::vector<uint8_t> depth(640*480*4);
	static std::vector<uint8_t> rgb(640*480*4);
	
pixBlock *videoKINECT :: getFrame()
{
	
  int size = m_image.image.xsize * m_image.image.ysize * m_image.image.csize;
  
  if (channel==0) //RGB OUTPUT -> Im sure there are better ways to convert from RGB to RGBA...
  {
		device->getRGB(rgb);
		
		if ((video_format==0) || (video_format==5))
		{
			int i=0;
				while (i<=size-1) {
					int num=(i%4)+(i/4)*3;
					if ((i % 4)==3)
						{
										m_image.image.data[i]=1.0; // set alpha to 1 - NOT WORKIN?!
						} else {
										m_image.image.data[i]=rgb[num];
						}
					i++;
				}
			} else if ((video_format==1) || (video_format==2)) {
				int i=0;
				while (i<=size-1) {
					int num=i/4;
					if ((i % 4)==3)
						{
										m_image.image.data[i]=1.0; // set alpha to 1 - NOT WORKIN?!
						} else {
										m_image.image.data[i]=rgb[num];
						}
					i++;
				}
			} else if ((video_format==4)) {
				int i=0;
				while (i<=size-1) {
					int num=i;
					m_image.image.data[i]=rgb[num];
					i++;
				}
		} else if (video_format==6) { //YUV RAW
				m_image.image.data=&rgb[0];
			}
		
			//m_image.image.fromRGB(&rgb);
	}
	
	if (channel==1) // DEPTH OUTPUT
	{
		device->getDepth(depth);
		m_image.image.data = &depth[0];
		
		/* // FIXED RGBA IN CALLBACK
			int i=0;
			while (i<=size-1) {
				int num=(i%4)+(i/4)*3;
				if ((i % 4)==3)
				{
								m_image.image.data[i]=1.0; //set alpha to 1
				} else {
								m_image.image.data[i]=depth[num];
				}
				i++;
			}
			*/
	}

  m_image.newimage = 1;
  m_image.image.notowned = true;
  m_image.image.upsidedown=true;
  return &m_image;
}

////////////////////////////////////////////////////////
// openDevice
//
/////////////////////////////////////////////////////////


bool videoKINECT :: openDevice(gem::Properties&props){
	
	if (device = &freenect.createDevice<MyFreenectDevice>(0))
	{
		verbose(1, "KINECT: Successfully opened...");
	} else {
		post("Could not open Kinect!");
	}
  return true;
}
////////////////////////////////////////////////////////
// closeDevice
//
/////////////////////////////////////////////////////////
void videoKINECT :: closeDevice(void){


}


/////////////////////////////////////////////////////////
// restartTransfer
//
/////////////////////////////////////////////////////////
bool videoKINECT :: restartTransfer()
{
  bool rendering=m_rendering;
  post("kinect: restart transfer");
  if(m_capturing)stopTransfer();
  post("kinect: restart stopped");
  if (rendering)startTransfer();
  post("kinect: restart started");

  return true;
}

////////////////////////////////////////////////////////
// startTransfer
//
/////////////////////////////////////////////////////////


bool videoKINECT :: startTransfer()
{

  m_image.image.xsize = m_width;
  m_image.image.ysize = m_height;
  //m_image.image.csize=3;
  m_image.image.setCsizeByFormat(m_reqFormat);
  m_image.image.reallocate();
  
  m_rendering=true;


  device->startVideo();
	device->startDepth();

	device->setVideoFormat(video_format);
	device->setDepthFormat(depth_format);
	
  return true;
}

/////////////////////////////////////////////////////////
// stopTransfer
//
/////////////////////////////////////////////////////////
bool videoKINECT :: stopTransfer()
{

	device->stopVideo();
	device->stopDepth();
	
  return true;
}

std::vector<std::string>videoKINECT :: enumerate(){
  std::vector<std::string>result;

  return result;
}

bool videoKINECT :: setColor(int format)
{
  if (format<=0 || format==m_reqFormat)return -1;
  m_reqFormat=format;
  restartTransfer();
  return true;
}

bool videoKINECT::enumProperties(gem::Properties&readable,
			      gem::Properties&writeable) {

	int i=0;
  std::vector<std::string>keys;
  gem::any type;

  readable.clear();
  writeable.clear();

  keys.clear();
  keys.push_back("angle");
  keys.push_back("led");

  keys.push_back("video_format");
  keys.push_back("depth_format");
  
  keys.push_back("channel");

  type=0;
  for(i=0; i<keys.size(); i++) {
    readable .set(keys[i], type);
    writeable.set(keys[i], type);
  }


}
void videoKINECT::getProperties(gem::Properties&props) {

  std::vector<std::string>keys=props.keys();
  int i;
  
  for(i=0; i<keys.size(); i++) {
    const std::string key=keys[i];

    if(key=="angle") {
			props.set(key, x_angle);
    } else if(key=="led") {
			props.set(key, x_led);
    } else if(key=="video_format") {
			props.set(key, (int)video_format);
    } else if(key=="depth_format") {
			props.set(key, (int)depth_format);
    } else if(key=="channel") {
			props.set(key, channel);
    } else if(key=="acc") {
			
			double dx,dy,dz;
			device->getState().getAccelerometers(&dx,&dy,&dz); // doesn't work in the moment
			post("acc %f %f %f",dx,dy,dz);
			props.set("acc_x", dx);
			props.set("acc_y", dy);
			props.set("acc_z", dz);
    }
  } /* key loop */
		
}

void videoKINECT::setProperties(gem::Properties&props) {
  std::vector<std::string>keys=props.keys();
  bool restart=false;
  bool do_s_chan=false, do_s_pict=false;
  int i=0;
  double d;
  std::string s;


  for(i=0; i<keys.size(); i++) {
    const std::string key=keys[i];
    if(0){;
#define RESTART_WITH_CHANGED_UINT(x) do { unsigned int ui=d; restart=(x!=ui); x=ui; } while(0)
    } else if (key=="angle") {
      if(props.get(key, d)) {
	if ( d<-30.0 ) { 
		x_angle = -30;
		post("not less than -30deg");
	} else if ( d>30.0 ) { 
		x_angle = 30;
		post("not more than -30deg");
	} else {
		x_angle=d;
	}
	device->setTiltDegrees(x_angle);
      }
    } else if (key=="led") {
      if(props.get(key, d)) {
	int led=d;
  if ( ( (int)led>=0 ) && ( (int)led<=5 ) ) x_led = (int)led;
  	if (x_led == 1) {
		device->setLed(LED_GREEN);
	}
	if (x_led == 2) {
		device->setLed(LED_RED);
	}
	if (x_led == 3) {
		device->setLed(LED_YELLOW);
	}
	if (x_led == 4) {
		device->setLed(LED_BLINK_GREEN);
	}
	if (x_led == 5) {
		device->setLed(LED_BLINK_RED_YELLOW);
	}
	if (x_led == 0) {
		device->setLed(LED_OFF);
	}
      }
    } else if (key=="video_format") {
      if(props.get(key, d)) {
					if ( ( (int)d>=0 ) && ( (int)d<=6 ) ) 
					{
						if (((int)d!=3) && ((int)d!=4)) {
							
							video_format = (freenect_video_format)(int)d;
							device->setVideoFormat(video_format);
							post("Video mode set to %i", (int)d);
						} else {
							post("Video mode %i currently not supported", (int)d);
						}
					} else {
						post("Not Valid Number! (0-6)");
				}
			}
    } else if (key=="depth_format") {
      if(props.get(key, d)) {
					if ( ( (int)d>=0 ) && ( (int)d<=3 ) ) 
					{
						if ( ((int)d!=2) && ((int)d!=3) )
						{
							depth_format = (freenect_depth_format)(int)d;
							device->setDepthFormat(depth_format);
							post("Depth mode set to %i", (int)d);
						} else {
							post("Depth mode %i currently not available", (int)d);
						}
						
					} else {
						post("Not Valid Number! (0-3)");
					}
			}
		} else if (key=="channel") {
			if(props.get(key, d)) {
					if ((int)d==0 ) 
					{
						// 0 -> RGB
						channel = (int)d;
						post("Channel set to RGB Output(%i)", (int)d);
					} else if ((int)d==1) {
						// 1-> Depth
						channel = (int)d;
						post("Channel set to Depth Output (%i)", (int)d);
					} else {
						post("Not Valid Number! (0 (RGB) or 1 (Depth))");
					}
			}
			
		}

	}
}

