/*-----------------------------------------------------------------

    GEM - Graphics Environment for Multimedia

    Load an video into a pix block

    Copyright (c) 1997-1999 Mark Danks. mark@danks.org
    Copyright (c) Günther Geiger. geiger@epy.co.at
    Copyright (c) 2001-2002 IOhannes m zmoelnig. forum::für::umläute. IEM. zmoelnig@iem.kug.ac.at
    For information on usage and redistribution, and for a DISCLAIMER OF ALL
    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.

    Linux version by Miller Puckette. msp@ucsd.edu
	
-----------------------------------------------------------------*/

#ifndef INCLUDE_videoKINECT_H_
#define INCLUDE_videoKINECT_H_
#include "plugins/video.h"

#include <assert.h>
// #include "libfreenect.h"
 #include "libfreenect.hpp"

#ifdef HAVE_PTHREAD
# include <pthread.h>
#endif

#include <stdio.h>
#include <iostream>
#include <string.h>
#include <cmath>
#include <vector>

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
	pix_video
    
    Loads in a video
    
KEYWORDS
    pix
    
DESCRIPTION

-----------------------------------------------------------------*/
namespace gem { class GEM_EXPORT videoKINECT : public video {
    public:
        //////////
        // Constructor
    	videoKINECT(void);
    	    	
    	//////////
    	// Destructor
    	virtual ~videoKINECT();

	////////
	// open the video-device
	virtual bool           openDevice(gem::Properties&);
	virtual void          closeDevice(void);
    
    	//////////
    	// Start up the video device
    	// [out] int - returns 0 if bad
    	bool	    	startTransfer(void);
	//////////
    	// Stop the video device
    	// [out] int - returns 0 if bad
    	bool	   	stopTransfer(void);

  //////////////////
  // restart the transfer if it is currently running
  virtual bool          restartTransfer(void);
  
  
	//////////
	// get the next frame
//	bool    grabFrame();
  virtual pixBlock    *getFrame();

	//////////
	// Set the video dimensions
	virtual bool	    	setColor(int);
	virtual bool        enumProperties(gem::Properties&, gem::Properties&);
	virtual void        getProperties(gem::Properties&);
	virtual void        setProperties(gem::Properties&);

	//////////
	// get available devices
	virtual std::vector<std::string>enumerate(void);
  
 protected:

  freenect_context *f_ctx;
  freenect_device *f_dev;
  
	int x_angle;
	int x_led;
	int x_dim;
  
  int channel; //0 -> RGB Out, 1-> Depth Out
  
	freenect_video_format video_format;
	freenect_depth_format depth_format;

  bool      m_rendering; // "true" when rendering is on, false otherwise


};
};

#endif	// for header file
